

set -ex



pip check
export ANACONDA_IDENT_DEBUG=1
export PYTHONUNBUFFERED=1
conda create -n testchild1 --yes
conda create -n testchild2 --yes
conda info
conda info --envs
anaconda-ident --status 2>&1
anaconda-keymgr --help 2>&1
python tests/test_config.py 2>&1
exit 0
